# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoGrant.ipynb.

# %% auto 0
__all__ = ['DomoGrant', 'DomoGrants']

# %% ../../nbs/classes/50_DomoGrant.ipynb 2
from dataclasses import dataclass, field
from typing import List

import httpx

import domolibrary.client.DomoAuth as dmda
import domolibrary.utils.DictDot as util_dd
import domolibrary.routes.grant as grant_routes

# %% ../../nbs/classes/50_DomoGrant.ipynb 5
@dataclass
class DomoGrant:
    id: str
    display_group: str = None
    title: str = None
    depends_on_ls: list[str] = None
    description: str = None
    role_membership_ls: list[str] = field(default=None)

    def __post_init__(self):
        self.id = str(self.id)

    def __eq__(self, other):
        if not isinstance(other, DomoGrant):
            return False

        return self.id == other.id

    @classmethod
    def _from_json(cls, obj):

        dd = obj
        if not isinstance(dd, util_dd.DictDot):
            dd = util_dd.DictDot(obj)

        return cls(
            id=dd.authority,
            display_group=dd.authorityUIGroup,
            depends_on_ls=dd.dependsOnAuthorities,
            title=dd.title,
            description=dd.description,
            role_membership_ls=[str(role) for role in dd.roleIds],
        )

# %% ../../nbs/classes/50_DomoGrant.ipynb 6
@dataclass
class DomoGrants:
    auth : dmda.DomoAuth = field(repr = False)

    grants : List[DomoGrant] = field(default = None)

    async def get(
        self,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        return_raw: bool = False,
    ):
        res = await grant_routes.get_grants(auth=self.auth, debug_api=debug_api, session=session)

        if return_raw or not res.is_success:
            return res

        self.grants = [DomoGrant._from_json(obj) for obj in res.response]

        return self.grants



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccount_Credential.ipynb.

# %% auto 0
__all__ = ['DAC_NoTargetInstance', 'DAC_NoTargetUser', 'DAC_NoPassword', 'DAC_NoUserName', 'DAC_NoAccessToken', 'DAC_ValidAuth',
           'DomoAccount_Credential']

# %% ../../nbs/classes/50_DomoAccount_Credential.ipynb 3
from typing import Callable

from dataclasses import dataclass, field
import httpx

import domolibrary.client.DomoError as dmde
import domolibrary.client.DomoAuth as dmda


import domolibrary.classes.DomoAccount_Default as dmacb
import domolibrary.classes.DomoAccessToken as dmact
import domolibrary.classes.DomoUser as dmdu
import domolibrary.utils.convert as dmcv

from nbdev.showdoc import patch_to

# %% ../../nbs/classes/50_DomoAccount_Credential.ipynb 5
class DAC_NoTargetInstance(dmde.ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no target_instance on class - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DAC_NoTargetUser(dmde.ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no target_user on class - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DAC_NoPassword(dmde.ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no password stored in account - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DAC_NoUserName(dmde.ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no username stored in account - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DAC_NoAccessToken(dmde.ClassError):
    def __init__(self, cls_instance):
        super().__init__(
            message=f"no access_token stored in account - {cls_instance.name}",
            cls_instance=cls_instance,
        )


class DAC_ValidAuth(dmde.ClassError):
    def __init__(self, cls_instance, message=None):
        super().__init__(
            message=message
            or f"{cls_instance.name} no valid auth retrieved for domo_instance - {cls_instance.target_instance}",
            cls_instance=cls_instance,
        )

# %% ../../nbs/classes/50_DomoAccount_Credential.ipynb 6
@dataclass
class DomoAccount_Credential(dmacb.DomoAccount_Default):

    target_instance: str = None

    is_valid_full_auth: bool = None
    is_valid_token_auth: bool = None

    _token_auth: dmda.DomoAuth = field(repr=False, default=None)
    _full_auth: dmda.DomoAuth = field(repr=False, default=None)

    target_auth: dmda.DomoAuth = field(default=None)
    target_user: dmdu.DomoUser = field(default=None)

    # def __post_init__(self):
    #     if not self.target_instance:
    #         raise DAC_NoTargetInstance(self)

    @classmethod
    def _class_from_json(
        cls,
        obj: dict,
        is_admin_summary: bool = True,
        auth: dmda.DomoAuth = None,
        **kwargs,
    ):
        return cls._default_from_json(
            obj=obj,
            is_admin_summary=is_admin_summary,
            auth=auth,
            target_instance=kwargs.get("target_instance"),
        )

    def set_password(self, password: str):
        self.Config.password = password
        return True

    def set_username(self, username: str):
        self.Config.username = username
        return True

    def set_access_token(self, access_token: str):
        self.Config.domo_access_token = access_token
        return True

    async def test_full_auth(
        self, debug_api: bool = False, session: httpx.AsyncClient = None
    ):
        """
        1. generates full auth object
        2. tests full auth object
        """
        self.is_valid_full_auth = False

        if not self.Config.username:
            raise DAC_NoUserName(self)

        if not self.Config.password:
            raise DAC_NoPassword(self)

        if not self.target_instance:
            raise DAC_NoTargetInstance(self)

        self._full_auth = dmda.DomoFullAuth(
            domo_instance=self.target_instance,
            domo_username=self.Config.username,
            domo_password=self.Config.password,
        )

        try:
            await self._full_auth.print_is_token(debug_api=debug_api, session=session)
            self.is_valid_full_auth = True

        except dmda.AuthError as e:
            dmcv.print_md(f"🤯 test_full_auth for: ***{self.name}*** returned {e}")

            self.is_valid_full_auth = False

        return self.is_valid_full_auth

    async def test_token_auth(
        self, debug_api: bool = False, session: httpx.AsyncClient = None
    ):
        """
        1. generates token auth object
        2. tests token auth object
        """

        self.is_valid_token_auth = False

        if not self.Config.domo_access_token:
            raise DAC_NoAccessToken(self)

        if not self.target_instance:
            raise DAC_NoTargetInstance(self)

        self._token_auth = dmda.DomoTokenAuth(
            domo_instance=self.target_instance,
            domo_access_token=self.Config.domo_access_token,
        )

        try:
            await self._token_auth.print_is_token(debug_api=debug_api, session=session)
            self.is_valid_token_auth = True
            self.target_auth = self._token_auth

        except dmda.AuthError as e:
            dmcv.print_md(f"🤯 test_token_auth for: ***{self.name}*** returned {e}")
            self.is_valid_token_auth = False

        return self.is_valid_token_auth

    def _set_target_auth(
        self,
        valid_backup_auth: dmda.DomoAuth = None,  # validated backup_auth
    ):
        """
        generates an auth object using the best of Token Auth < Full Auth
        uses a validated backup_auth as a failover target_auth
        """

        target_auth = None

        if self.is_valid_token_auth:
            target_auth = self._token_auth

        if self.is_valid_full_auth:
            target_auth = self._full_auth

        if not target_auth and (valid_backup_auth and valid_backup_auth.is_valid_token):
            target_auth = valid_backup_auth

        if not target_auth:
            raise DAC_ValidAuth(self)

        self.target_auth = target_auth

        return self.target_auth

    async def test_auths(
        self,
        backup_auth: dmda.DomoAuth = None,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
    ):
        ## test token auth
        try:
            await self.test_token_auth(debug_api=debug_api, session=session)

        except dmde.DomoError as e:
            print(f"testing token: {self.name}: {e}")

        ## test full auth
        try:
            await self.test_full_auth(debug_api=debug_api, session=session)

        except dmde.DomoError as e:
            print(f"testing full auth: {self.name}: {e}")

        ## generate target_auth
        try:
            self._set_target_auth(valid_backup_auth=backup_auth)

        except DAC_ValidAuth as e:
            print(f"{self.name}: unable to generate valid target_auth: {e}")

        return self.to_json()

    
    def to_json(self):
        return {
            "account_id": self.id,
            "alias": self.name,
            "target_instance": self.target_instance,
            "is_valid_full_auth": self.is_valid_full_auth,
            "is_valid_token_auth": self.is_valid_token_auth,
        }

# %% ../../nbs/classes/50_DomoAccount_Credential.ipynb 7
@patch_to(DomoAccount_Credential)
async def get_target_user(
    self,
    user_email: None,  # defaults to username from the AccountConfig object
    backup_auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    is_test_auths: bool = False,
):
    user_email = self.Config.username or user_email

    if not user_email:
        raise DAC_NoUserName(self)

    if is_test_auths or not self.target_auth:
        self.test_auths(session=session, debug_api=debug_api, backup_auth=backup_auth)

    if not self.target_auth:
        raise DAC_ValidAuth(
            self,
            message="no target_auth - run with is_test_auths = True and/or pass a valid backup_auth",
        )

    self.target_user = await dmdu.DomoUsers.by_email(
        email_ls=[user_email],
        auth=self.target_auth,
        debug_api=debug_api,
        session=session,
    )

    return self.target_user


@patch_to(DomoAccount_Credential)
async def update_target_user_password(
    self: DomoAccount_Credential,
    new_password: str,
    user_email: str = None,  # defaults to username from the AccountConfig object
    is_test_auths: bool = False,
    is_update_account: bool = True,
    backup_auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):

    if is_test_auths or not self.target_auth:
        self.test_auths(session=session, debug_api=debug_api, backup_auth=backup_auth)

    if not self.target_user:
        await self.get_target_user(
            debug_api=debug_api,
            session=session,
            user_email=user_email,
            is_test_auths=False,
        )

    await self.target_user.reset_password(
        new_password=new_password, debug_api=debug_api, session=session
    )

    self.set_password(new_password)

    if is_update_account:
        await self.update_config(debug_api=debug_api, session=session)

    return self


@patch_to(DomoAccount_Credential)
async def reset_access_token(
    self,
    token_name=None,
    duration_in_days=90,
    user_email: str = None,  # defaults to username from the AccountConfig object
    is_test_auths: bool = False,
    is_update_account: bool = True,
    backup_auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):

    if is_test_auths or not self.target_auth:
        self.test_auths(session=session, debug_api=debug_api, backup_auth=backup_auth)

    if not self.target_user:
        await self.get_target_user(
            debug_api=debug_api,
            session=session,
            user_email=user_email,
            is_test_auths=False,
        )

    if not token_name:
        token_name = self.target_user.email_address

    domo_access_tokens = await self.target_user.get_access_tokens(
        session=session, debug_api=debug_api
    )

    domo_access_token = next(
        (dat for dat in domo_access_tokens if dat.name.lower() == token_name.lower()),
        None,
    )

    if domo_access_token:
        await domo_access_token.regenerate(
            duration_in_days=duration_in_days, session=session, debug_api=debug_api
        )

    else:
        domo_access_token = await dmact.DomoAccessToken.generate(
            duration_in_days=duration_in_days,
            token_name=token_name,
            auth=self.target_auth,
            owner=self.target_user,
            debug_api=debug_api,
            session=session,
        )

    self.set_access_token(domo_access_token.token)

    if is_update_account:
        await self.update_config(debug_api=debug_api, session = session)

    return self



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccess.ipynb.

# %% auto 0
__all__ = ['Access_Entity', 'Access_Config_Error', 'DomoAccess', 'DomoAccess_Account', 'DomoAccess_OAuth']

# %% ../../nbs/classes/50_DomoAccess.ipynb 3
from domolibrary.routes.account import (
    ShareAccount_AccessLevel,
    ShareAccount_V1_AccessLevel,
    ShareAccount,
    Account_Share_Error,
)

# %% ../../nbs/classes/50_DomoAccess.ipynb 4
from dataclasses import dataclass, field
from typing import Any, List

from abc import abstractmethod

import httpx

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde
import domolibrary.routes.account as account_routes

import domolibrary.utils.chunk_execution as dmce

# %% ../../nbs/classes/50_DomoAccess.ipynb 7
@dataclass
class Access_Entity:
    entity: Any
    access_level: ShareAccount
    auth: dmda.DomoAuth = field(repr=False)

    @classmethod
    async def from_user_id(
        cls,
        user_id,
        access_level: ShareAccount,
        auth: dmda.DomoAuth,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
    ):

        # if entity_type == "USER":
        import domolibrary.classes.DomoUser as dmdu

        domo_user = await dmdu.DomoUser.get_by_id(
            user_id=user_id,
            auth=auth,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        return cls(entity=domo_user, access_level=access_level, auth=auth)

    @classmethod
    async def from_group_id(
        cls,
        group_id,
        access_level: ShareAccount,
        auth: dmda.DomoAuth,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
    ):

        import domolibrary.classes.DomoGroup as dmdg

        domo_group = await dmdg.DomoGroup.get_by_id(
            group_id=group_id,
            auth=auth,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        return cls(entity=domo_group, access_level=access_level, auth=auth)

    def __eq__(self, other):
        return (
            self.entity.id == other.entity.id
            and self.entity.__class__.__name__ == other.entity.__class__.__name__
        )

# %% ../../nbs/classes/50_DomoAccess.ipynb 8
class Access_Config_Error(dmde.ClassError):
    def __init__(self, cls_instance, account_id, message):
        super().__init__(
            cls_instance=cls_instance, message=message, entity_id=account_id
        )


@dataclass
class DomoAccess:
    auth: dmda.DomoAuth = field(repr=False)

    share_enum: Any = field(repr = False)

    accesslist: List[Access_Entity] = field(default_factory=lambda: []) # can include users or groups

    accesslist_all_users: List[Any] = field(default_factory=lambda: [])  # DomoUser

    parent_id: int = None
    parent: any = field(repr=False, default=None)

    @abstractmethod
    async def get(self, *args, **kwargs):
        pass

    def _common_init(self):
        if not self.parent_id and self.parent and hasattr(self.parent, "id"):
            self.parent_id = self.parent.id

    def __post_init__(self):
        self._common_init()

    async def get_all_users(
        self,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 1,
        session: httpx.AsyncClient = None,
    ):

        import domolibrary.classes.DomoUser as dmdu
        import domolibrary.classes.DomoGroup as dmdg

        await self.get(
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        for access in self.accesslist:
            entity = access.entity

            if isinstance(entity, dmdu.DomoUser):
                if entity not in self.accesslist_all_users:
                    self.accesslist_all_users.append(entity)

            if isinstance(entity, dmdg.DomoGroup):
                domo_users = await access.entity.Membership.get_members()

                [
                    self.accesslist_all_users.append(user)
                    for user in domo_users
                    if user not in self.accesslist_all_users
                ]

        return self.accesslist_all_users

    async def share(
        self,
        user_id=None,
        group_id=None,
        domo_group=None,
        domo_user=None,
        access_level: ShareAccount = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        access_level = access_level or self.share_enum.default

        share_payload = access_level.generate_payload(
            user_id=user_id or (domo_user and domo_user.id),
            group_id=group_id or (domo_group and domo_group.id),
        )

        res = await access_level.share(
            auth=self.auth,
            account_id=self.parent_id,
            share_payload=share_payload,
            debug_api=debug_api,
            session=session,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        return res
    
    async def upsert_share(
        self,
        user_id=None,
        group_id=None,
        domo_group=None,
        domo_user=None,
        access_level: ShareAccount = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        user_id=user_id or (domo_user and domo_user.id)
        group_id=group_id or (domo_group and domo_group.id)

        await self.get(session = session, debug_api = debug_api, debug_num_stacks_to_drop = debug_num_stacks_to_drop + 1, )
        
        for member in self.accesslist:
            if ( member.entity.id == user_id or member.entity.id == group_id) and access_level == member.access_level:
                return False

        return await self.share(
            user_id = user_id,
            group_id = group_id,
            access_level= access_level,
            debug_api= debug_api,
            debug_num_stacks_to_drop= debug_num_stacks_to_drop,
            session = session
        )    
        

# %% ../../nbs/classes/50_DomoAccess.ipynb 10
@dataclass
class DomoAccess_Account(DomoAccess):
    version: int = None  # api version - aligns to feature switch

    share_enum: ShareAccount = field(repr = False, default = ShareAccount_AccessLevel)

    def __post_init__(self):
        self._common_init()
        
        if isinstance(self.share_enum , ShareAccount_AccessLevel):
            self.version = 2
        elif isinstance(self.share_enum , ShareAccount_V1_AccessLevel):
            self.version = 1
        
        return True

    async def get(
        self,
        debug_api: bool = False,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
    ):
        """assumes v2 access api"""

        res = await account_routes.get_account_accesslist(
            auth=self.auth,
            account_id=self.parent_id,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.accesslist = await dmce.gather_with_concurrency(
            *[
                Access_Entity.from_user_id(
                    user_id=obj["id"],
                    auth=self.auth,
                    access_level=self.share_enum(obj["accessLevel"]),
                )
                for obj in res.response
                if obj["type"] == "USER"
            ],
            *[
                Access_Entity.from_group_id(
                    group_id=obj["id"],
                    auth=self.auth,
                    access_level=self.share_enum(obj["accessLevel"]),
                )
                for obj in res.response
                if obj["type"] == "GROUP"
            ],
            n=10,
        )
        return self.accesslist

# %% ../../nbs/classes/50_DomoAccess.ipynb 18
@dataclass
class DomoAccess_OAuth(DomoAccess):
    share_enum: account_routes.ShareAccount = field(
        repr=False, default=ShareAccount_AccessLevel
    )

    async def get(
        self,
        debug_api: bool = False,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
    ):
        """assumes v2 access api"""

        res = await account_routes.get_oauth_account_accesslist(
            auth=self.auth,
            account_id=self.parent_id,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.accesslist = await dmce.gather_with_concurrency(
            *[
                Access_Entity.from_user_id(
                    user_id=obj["id"],
                    auth=self.auth,
                    access_level=self.share_enum(obj["accessLevel"]),
                )
                for obj in res.response
                if obj["type"] == "USER"
            ],
            *[
                Access_Entity.from_group_id(
                    group_id=obj["id"],
                    auth=self.auth,
                    access_level=self.share_enum(obj["accessLevel"]),
                )
                for obj in res.response
                if obj["type"] == "GROUP"
            ],
            n=10,
        )
        return self.accesslist



"""a class based approach for interacting with Domo Datasets"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoDataset_Schema.ipynb.

# %% auto 0
__all__ = ['DatasetSchema_Types', 'DomoDataset_Schema_Column', 'DomoDataset_Schema', 'DatasetSchema_InvalidSchema']

# %% ../../nbs/classes/50_DomoDataset_Schema.ipynb 3
from dataclasses import dataclass, field
from typing import List, Any
from enum import Enum

from nbdev.showdoc import patch_to

import domolibrary.utils.DictDot as util_dd

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde
import pandas as pd

import domolibrary.routes.dataset as dataset_routes

# %% ../../nbs/classes/50_DomoDataset_Schema.ipynb 7
class DatasetSchema_Types(Enum):
    STRING = "STRING"
    DOUBLE = "DOUBLE"
    LONG = "LONG"
    DATE = "DATE"
    DATETIME = "DATETIME"


@dataclass
class DomoDataset_Schema_Column:
    name: str
    id: str
    type: DatasetSchema_Types
    order: int = 0
    visible: bool = True
    upsert_key: bool = False

    def __eq__(self, other):
        return self.id == other.id

    @classmethod
    def _from_json(cls, json_obj):
        dd = util_dd.DictDot(json_obj)
        return cls(
            name=dd.name,
            id=dd.id,
            type=dd.type,
            visible=dd.visible or dd.isVisible or True,
            upsert_key=dd.upsertKey or False,
            order=dd.order or 0,
        )

    def to_dict(self):
        s = self.__dict__
        s["upsertKey"] = s.pop("upsert_key") if "upsert_key" in s else False
        return s




# %% ../../nbs/classes/50_DomoDataset_Schema.ipynb 9
@dataclass
class DomoDataset_Schema:
    """class for interacting with dataset schemas"""

    auth : dmda.DomoAuth = field(repr = False)
    dataset_id : str 

    parent: Any = field(repr = False, default = None)
    columns: List[DomoDataset_Schema_Column] = field(default_factory=list)

    def __post_init__(self):
        if self.parent: 
            self.auth = self.parent.auth
            self.dataset_id = self.parent.id


    @classmethod
    def _from_parent(cls, parent):
        ds = cls(
            parent = parent,
            auth = parent.auth,
            dataset_id = parent.id
        )

        parent.Schema = ds
        return ds

    async def get(
        self,
        debug_api: bool = False,
        return_raw: bool = False,  # return the raw response
    ) -> List[DomoDataset_Schema_Column]:
        """method that retrieves schema for a dataset"""


        res = await dataset_routes.get_schema(
            auth=self.auth, dataset_id=self.dataset_id, debug_api=debug_api
        )

        if return_raw:
            return res.response

        self.columns = [
            DomoDataset_Schema_Column._from_json(json_obj=obj)
            for obj in res.response.get("tables")[0].get("columns")
        ]

        return self.columns

# %% ../../nbs/classes/50_DomoDataset_Schema.ipynb 11
class DatasetSchema_InvalidSchema(dmde.DomoError):
    def __init__(self, domo_instance, dataset_id, missing_columns, dataset_name=None, message = None):

        if dataset_id:
            message = f"{dataset_id}{f' - {dataset_name}' if dataset_name else ''} is missing columns {', '.join(missing_columns)}"

        super().__init__(domo_instance=domo_instance, message=message)


@patch_to(DomoDataset_Schema)
async def _test_missing_columns(
    self: DomoDataset_Schema,
    df: pd.DataFrame,
    dataset_id=None,
    auth: dmda.DomoAuth = None,
):

    dataset_id = dataset_id or self.parent.id
    auth = auth or self.parent.auth

    await self.get()

    missing_columns = [
        col for col in df.columns if col not in [scol.name for scol in self.columns]
    ]

    if len(missing_columns) > 0:
        raise DatasetSchema_InvalidSchema(
            domo_instance=auth.domo_instance,
            dataset_id=dataset_id,
            missing_columns=missing_columns,
        )

    return False

# %% ../../nbs/classes/50_DomoDataset_Schema.ipynb 14
@patch_to(DomoDataset_Schema)
async def reset_col_order(self: DomoDataset_Schema, df: pd.DataFrame):

    await self.get()

    if len(self.columns) != len(df.columns):
        raise Exception("")

    for index, col in enumerate(self.schema.columns):
        col.order = col.order if col.order > 0 else index

    # for index, schema in enumerate(consol_ds.schema.columns):
    #     schema.order = index

    # schema = {'columns': [col.__dict__ for col in consol_ds.schema.columns]}
    # schema

    # import domolibrary.routes.dataset as dataset_routes
    # await dataset_routes.alter_schema(auth = consol_auth, dataset_id = consol_ds.id, schema_obj = schema)

# %% ../../nbs/classes/50_DomoDataset_Schema.ipynb 15
@patch_to(DomoDataset_Schema)
def to_dict(self: DomoDataset_Schema):
    return {"columns": [col.to_dict() for col in self.columns]}

# %% ../../nbs/classes/50_DomoDataset_Schema.ipynb 17
@patch_to(DomoDataset_Schema)
def add_col(
    self: DomoDataset_Schema, col: DomoDataset_Schema_Column, debug_prn: bool = False
):

    if col in self.columns and debug_prn:
        print(
            f"column - {col.name} already in dataset {self.parent.name if self.parent else '' }"
        )

    if col not in self.columns:
        self.columns.append(col)

    return self.columns


@patch_to(DomoDataset_Schema)
def remove_col(
    self: DomoDataset_Schema,
    col_to_remove: DomoDataset_Schema_Column,
    debug_prn: bool = False,
):

    [
        self.columns.pop(index)
        for index, col in enumerate(self.columns)
        if col == col_to_remove
    ]

    return self.columns

# %% ../../nbs/classes/50_DomoDataset_Schema.ipynb 19
class CRUD_Dataset_Error(dmde.DomoError):
    def __init__(self, auth, res, message):
        super().__init__(
            status = res.status,
            function_name= res.traceback_details.function_name,
            parent_class = res.traceback_details.parent_class,
            message = message or res.response,
            domo_instance = auth.domo_instance
        )

@patch_to(DomoDataset_Schema)
async def alter_schema(
    self: DomoDataset_Schema,
    dataset_id: str = None,
    auth: dmda.DomoAuth = None,
    return_raw: bool = False,
    debug_api: bool = False,
):

    dataset_id = dataset_id or self.parent.id
    auth = auth or self.parent.auth

    schema_obj = self.to_dict()

    if return_raw:
        return schema_obj

    res =  await dataset_routes.alter_schema(
        dataset_id=dataset_id, auth=auth, schema_obj=schema_obj, debug_api=debug_api
    )

    if not res.is_success:
        raise CRUD_Dataset_Error(auth = auth , res = res, message = f"unable to alter schema for {dataset_id}")

    return res
